"""
Implements the Rowshandel (2013) directivity model. Note that the report
provides many options for computing the directivity factor and we have
also used some coefficients that are unpublished updates by Rowshandel.
Some of the implementation options are controlled by arguments (e.g.,
mtype).

Fd is the directivity function. One of the options is whether or not to
use the 'centering' term (the argument is 'centered' and defaults to
True). If it is True then

Fd = C1 * (XiPrime - Xic) * LD * DT * WP

otherwise

Fd = (C1 * Xiprime + C2) * LD * DT * WP

where

- C1 and C2 are regression coefficients.
- XiPrime is a factor that accounts for rupture propagation and slip
  direction.
- Xic is the centering term.
- LD is the rupture length de-normalization factor.
- DT is the distance-taper.
- WP is the narrow-band multiplier.

Note that Fd is intended to be used in GMPEs as:

ln(IM_dir) = ln(IM) + Fd

where

- Fd is the directivity effect.
- IM is the intensity measure predicted by the GMPE.
- IM_dir is the directivity-adjusted IM.

To do
    - Add checks on function arguments (e.g., mtype) for valid values.
    - Interpolate periods.

References:
    Rowshandel, B. (2013). Rowshandel’s NGA-West2 directivity model,
    Chapter 3 of PEER Report No. 2013/09, P. Spudich (Editor), Pacific
    Earthquake Engineering Research Center, Berkeley, CA.
    `[link] <http://peer.berkeley.edu/publications/peer_reports/reports_2013/webPEER-2013-09-Spudich.pdf>`__
"""  # noqa

# stdlib imports
import copy

# third party imports
import numpy as np
import openquake.hazardlib.geo as geo
from esi_utils_rupture import utils
from esi_utils_rupture.distance import get_distance
from esi_utils_vectors.ecef import ecef2latlon, latlon2ecef
from esi_utils_vectors.vector import Vector
from openquake.hazardlib.geo.utils import OrthographicProjection


class Rowshandel2013(object):

    __c1 = [0.35, 0.75, 0.95, 1.28, 1.60]
    __c2 = [-0.035, -0.10, -0.15, -0.26, -0.30]
    __periods = [1.0, 3.0, 5.0, 7.5, 10.0]

    def __init__(
        self,
        origin,
        rup,
        lat,
        lon,
        dep,
        dx,
        T,
        a_weight=0.5,
        mtype=1,
        simpleDT=False,
        centered=True,
    ):
        """
        Constructor for rowshandel2013.

        Args:
            origin: Origin instance.
            rup: Rupture instance.
            lat (ndarray): Numpy array of site latitudes.
            lon (ndarray): Numpy array of site longitudes.
            dep (ndarray): Numpy array of site depths (km); positive down.
            dx (float): Target mesh spacing for subruptures in km. The mesh
                snaps to the edges of the quadrilaterals so the actual mesh
                spacing will not equal this exactly; spacing in x and y will
                not be equal.
            T (list): List floats (or a float) for periods to compute fd;
                Currently, only acceptable values are 1, 2, 3, 4, 5, 7.5.
            a_weight (float): Weighting factor for how p-dot-q and s-dot-q are
                averaged; 0 for only p-dot-q (propagation factor) and 1 for
                only s-dot-q (slip factor).
            mtype (int): Either 1 or 2; 1 for adding only positive components
                of dot products, 2 for adding all components of dot products.
            simpleDT (bool): Should the simpler DT equation be used? Usually
                False.
            centered (bool): Should the centered directivity parameter be used?
        """
        self._origin = origin
        self._rup = rup
        self._hyp = origin.getHypo()
        self._lat = lat
        self._lon = lon
        self._dep = dep
        self._rake = origin.rake
        self._dx = dx
        self._M = origin.mag
        if isinstance(T, float):
            self._T = [T]
        else:
            self._T = T
        if (a_weight > 1.0) or (a_weight < 0):
            raise ValueError("a_weight must be between 0 and 1.")
        self._a_weight = a_weight
        if (mtype != 1) and (mtype != 2):
            raise ValueError("mtype can onlty be 1 or 2.")
        self._mtype = mtype
        self._simpleDT = simpleDT
        self._centered = centered

        # Period independent parameters
        self.__computeWrup()
        self.__computeLD()
        self.__computeXiPrime()
        self.__getCenteringTerm()

        self.__computeFd()

    @classmethod
    def fromSites(
        cls,
        origin,
        rup,
        sites,
        dx,
        T,
        a_weight=0.5,
        mtype=1,
        simpleDT=False,
        centered=True,
    ):
        """Construct a rowshandel2013 instance from a sites instance.

        Class method for constructing a rowshandel2013 instance from
        a sites instance.

        Args:
            origin: Origin instance.
            rup: Rupture instance.
            sites: Sites object.
            dx: Float for target mesh spacing for subruptures in km. The mesh
                snaps to the edges of the quadrilaterals so the actual mesh
                spacing will not equal this exactly; spacing in x and y will
                not be equal.
            T: Period; Currently, only acceptable values are 1, 2, 3, 4, 5,
                7.5.
            a_weight: Weighting factor for how p-dot-q and s-dot-q are
                averaged; 0 for only p-dot-q (propagation factor) and 1 for
                only s-dot-q (slip factor).
            mtype: Integer, either 1 or 2; 1 for adding only positive
                components of dot products, 2 for adding all components of dot
                products.
            simpleDT: Boolean; should the simpler DT equation be used? Usually
                False.
            centered: Boolean; should the centered directivity parameter be
                used

        Returns:
            Rowshandel2013 directivity class.
        """
        sm_dict = sites.getVs30Grid().getGeoDict()
        west = sm_dict.xmin
        east = sm_dict.xmax
        south = sm_dict.ymin
        north = sm_dict.ymax
        nx = sm_dict.nx
        ny = sm_dict.ny
        lats = np.linspace(north, south, ny)
        lons = np.linspace(west, east, nx)
        lon, lat = np.meshgrid(lons, lats)
        dep = np.zeros_like(lon)
        return cls(
            origin, rup, lat, lon, dep, dx, T, a_weight, mtype, simpleDT, centered
        )

    def getFd(self):
        """
        :returns:
            Numpy array of Fd; this is the directivity amplification factor
            (log units).
        """
        return copy.deepcopy(self._fd)

    def getXiPrime(self):
        """
        Returns:
            Numpy array of Xi'; this is the variable that accounts for
            geometric factors in Fd.
        """
        return copy.deepcopy(self._xi_prime)

    def getDT(self):
        """
        Returns:
           List of numpy arrays of the distance taper factor. Length is the
           number of periods.
        """
        return copy.deepcopy(self._DTlist)

    def getWP(self):
        """
        Returns:
            List of numpy array of the narrow-band multiplier. Length is the
            number of periods.
        """
        return copy.deepcopy(self._WPlist)

    def getLD(self):
        """
        Returns:
            Numpy array of the rupture length de-normalization factor.
        """
        return copy.deepcopy(self._LD)

    def __computeFd(self):
        """
        Fd is the term that modifies the GMPE output as an additive term
        (in log space).
        ln(Y) = f(M, R, ...) + Fd
        """

        xcipc = self._xi_prime - self._xi_c

        # fd is a list with same length as T
        self._fd = [None] * len(self._T)
        self._DTlist = [None] * len(self._T)
        self._WPlist = [None] * len(self._T)
        for i, period in enumerate(self._T):
            c1sel = self.__c1[self.__periods == period]
            c2sel = self.__c2[self.__periods == period]

            # Period dependent parameters
            self.__computeWP(period)
            self.__computeDT(period)

            self._DTlist[i] = copy.deepcopy(self._DT)
            self._WPlist[i] = copy.deepcopy(self._WP)

            if self._centered:
                self._fd[i] = c1sel * self._WP * self._DT * xcipc
            else:
                self._fd[i] = (
                    (c1sel * self._xi_prime + c2sel) * self._LD * self._DT * self._WP
                )

    def __computeWrup(self):
        """
        Compute the the portion (in km) of the width of the rupture which
        ruptures up-dip from the hypocenter to the top of the rupture.

        Wrup is the portion (in km) of the width of the rupture which
        ruptures up-dip from the hypocenter to the top of the rupture.

        * This is ambiguous for ruptures with varible top of rupture (not
          allowed in NGA). For now, lets just compute this for the
          quad where the hypocenter is located.
        * Alternative is to compute max Wrup for the different quads.

        """
        nquad = len(self._rup.getQuadrilaterals())

        # ---------------------------------------------------------------------
        # First find which quad the hypocenter is on
        # ---------------------------------------------------------------------

        x, y, z = latlon2ecef(self._hyp.latitude, self._hyp.longitude, self._hyp.depth)
        hyp_ecef = np.array([[x, y, z]])
        qdist = np.zeros(nquad)
        for i in range(0, nquad):
            qdist[i] = utils._quad_distance(self._rup.getQuadrilaterals()[i], hyp_ecef)[
                0, 0
            ]
        ind = int(np.where(qdist == np.min(qdist))[0][0])
        # *** check that this doesn't break with more than one quad
        q = self._rup.getQuadrilaterals()[ind]

        # ---------------------------------------------------------------------
        # Compute Wrup on that quad
        # ---------------------------------------------------------------------

        pp0 = Vector.fromPoint(
            geo.point.Point(q[0].longitude, q[0].latitude, q[0].depth)
        )
        hyp_ecef = Vector.fromPoint(
            geo.point.Point(self._hyp.longitude, self._hyp.latitude, self._hyp.depth)
        )
        hp0 = hyp_ecef - pp0
        ddv = utils.get_quad_down_dip_vector(q)
        self._Wrup = Vector.dot(ddv, hp0) / 1000

    def __computeXiPrime(self):
        """
        Computes the xi' value.
        """
        hypo_ecef = Vector.fromPoint(
            geo.point.Point(self._hyp.longitude, self._hyp.latitude, self._hyp.depth)
        )

        slat = self._lat
        slon = self._lon

        # Convert site to ECEF:
        site_ecef_x = np.ones_like(slat)
        site_ecef_y = np.ones_like(slat)
        site_ecef_z = np.ones_like(slat)

        # Make a 3x(#number of sites) matrix of site locations
        # (rows are x, y, z) in ECEF
        site_ecef_x, site_ecef_y, site_ecef_z = latlon2ecef(
            slat, slon, np.zeros(slon.shape)
        )
        site_mat = np.array(
            [
                np.reshape(site_ecef_x, (-1,)),
                np.reshape(site_ecef_y, (-1,)),
                np.reshape(site_ecef_z, (-1,)),
            ]
        )

        xi_prime_unscaled = np.zeros_like(slat)

        # Normalize by total number of subruptures. For mtype == 1, the number
        # of subruptures will vary with site and be different for xi_s and
        # xi_p, so keep two variables and sum them for each quad.
        nsubs = np.zeros(np.prod(slat.shape))
        nsubp = np.zeros(np.prod(slat.shape))

        xi_prime_s = np.zeros(np.prod(slat.shape))
        xi_prime_p = np.zeros(np.prod(slat.shape))

        for k in range(len(self._rup.getQuadrilaterals())):
            # Select a quad
            q = self._rup.getQuadrilaterals()[k]

            # Quad mesh (ECEF coords)
            mesh = utils.get_quad_mesh(q, self._dx)

            # Rupture plane normal vector (ECEF coords)
            rpnv = utils.get_quad_normal(q)
            rpnvcol = np.array([[rpnv.x], [rpnv.y], [rpnv.z]])

            cp_mat = np.array(
                [
                    np.reshape(mesh["cpx"], (-1,)),
                    np.reshape(mesh["cpy"], (-1,)),
                    np.reshape(mesh["cpz"], (-1,)),
                ]
            )

            # Compute matrix of p vectors
            hypcol = np.array([[hypo_ecef.x], [hypo_ecef.y], [hypo_ecef.z]])
            pmat = cp_mat - hypcol

            # Project pmat onto quad
            ndotp = np.sum(pmat * rpnvcol, axis=0)
            pmat = pmat - ndotp * rpnvcol

            mag = np.sqrt(np.sum(pmat * pmat, axis=0))
            pmatnorm = pmat / mag  # like r1

            # According to Rowshandel:
            #   "The choice of the +/- sign in the above equations
            #    depends on the (along-the-strike and across-the-dip)
            #    location of the rupturing sub-fault relative to the
            #    location of the hypocenter."
            # and:
            #   "for the along the strike component of the slip unit
            #    vector, the choice of the sign should result in the
            #    slip unit vector (s) being exactly the same as  the
            #    rupture unit vector (p) for a pure strike-slip case"

            # Strike slip and dip slip components of unit slip vector
            # (ECEF coords)
            ds_mat, ss_mat = _get_quad_slip_ds_ss(q, self._rake, cp_mat, pmatnorm)

            slpmat = ds_mat + ss_mat
            mag = np.sqrt(np.sum(slpmat * slpmat, axis=0))
            slpmatnorm = slpmat / mag

            # Loop over sites
            for i in range(site_mat.shape[1]):
                sitecol = np.array(
                    [[site_mat[0, i]], [site_mat[1, i]], [site_mat[2, i]]]
                )

                qmat = sitecol - cp_mat  # 3x(ni*nj), like r2
                mag = np.sqrt(np.sum(qmat * qmat, axis=0))
                qmatnorm = qmat / mag

                # Propagation dot product
                pdotqraw = np.sum(pmatnorm * qmatnorm, axis=0)

                # Slip vector dot product
                sdotqraw = np.sum(slpmatnorm * qmatnorm, axis=0)

                if self._mtype == 1:
                    # Only sum over (+) directivity effect subruptures

                    # xi_p_prime
                    pdotq = pdotqraw.clip(min=0)
                    nsubp[i] = nsubp[i] + np.sum(pdotq > 0)

                    # xi_s_prime
                    sdotq = sdotqraw.clip(min=0)
                    nsubs[i] = nsubs[i] + np.sum(sdotq > 0)

                elif self._mtype == 2:
                    # Sum over contributing subruptures

                    # xi_p_prime
                    pdotq = pdotqraw
                    nsubp[i] = nsubp[i] + cp_mat.shape[1]

                    # xi_s_prime
                    sdotq = sdotqraw
                    nsubs[i] = nsubs[i] + cp_mat.shape[1]

                # Normalize by n sub ruptures later
                xi_prime_s[i] = xi_prime_s[i] + np.sum(sdotq)
                xi_prime_p[i] = xi_prime_p[i] + np.sum(pdotq)

        # Apply a water level to nsubp and nsubs to avoid division by
        # zero. This should only occur when the numerator is also zero
        # and so the resulting value should be zero.
        nsubs = np.maximum(nsubs, 1)
        nsubp = np.maximum(nsubp, 1)

        # We are outside the 'k' loop over nquads.
        # o Normalize xi_prime_s and xi_prime_p
        # o Reshape them
        # o Add them together with the 'a' weights
        xi_prime_tmp = (self._a_weight) * (xi_prime_s / nsubs) + (
            1 - self._a_weight
        ) * (xi_prime_p / nsubp)
        xi_prime_unscaled = xi_prime_unscaled + np.reshape(xi_prime_tmp, slat.shape)

        # Scale so that xi_prime has range (0, 1)
        if self._mtype == 1:
            xi_prime = xi_prime_unscaled
        elif self._mtype == 2:
            xi_prime = 0.5 * (xi_prime_unscaled + 1)

        self._xi_prime = xi_prime

    def __computeLD(self):
        """
        Computes LD -- the rupture length de-normalization term.
        """
        # Use an orthographic projection
        latmin = self._lat.min()
        latmax = self._lat.max()
        lonmin = self._lon.min()
        lonmax = self._lon.max()
        proj = OrthographicProjection(lonmin, lonmax, latmax, latmin)

        # Get epi projection
        epi_x, epi_y = proj(self._hyp.longitude, self._hyp.latitude)

        # Get the lines for the top edge of the rupture
        qds = self._rup.getQuadrilaterals()
        nq = len(qds)
        top_lat = np.array([[]])
        top_lon = np.array([[]])
        top_dep = np.array([[]])
        for j in range(0, nq):
            top_lat = np.append(top_lat, qds[j][0].latitude)
            top_lat = np.append(top_lat, qds[j][1].latitude)
            top_lon = np.append(top_lon, qds[j][0].longitude)
            top_lon = np.append(top_lon, qds[j][1].longitude)
            top_dep = np.append(top_dep, qds[j][0].depth)
            top_dep = np.append(top_dep, qds[j][1].depth)
        top_x, top_y = proj(top_lon, top_lat)

        Lrup_max = 400
        slat = self._lat
        slon = self._lon
        LD = np.zeros_like(slat)
        Ls = np.zeros_like(slat)

        ni, nj = LD.shape
        for i in range(ni):
            for j in range(nj):
                # -------------------------------------------------------------
                # Compute Ls
                # -------------------------------------------------------------
                # Convert to local orthographic
                site_x, site_y = proj(slon[i, j], slat[i, j])

                # Shift so center is at epicenter
                site_x2 = site_x - epi_x
                site_y2 = site_y - epi_y
                top_x2 = top_x - epi_x
                top_y2 = top_y - epi_y

                # Angle to rotate to put site on x-axis
                alpha = np.arctan2(site_y2, site_x2)
                axis = [0, 0, 1]
                rmat = _rotation_matrix(axis, -alpha)
                llr = [None] * len(top_lat)

                # Apply rotation to each point on trace
                for k in range(len(top_lat)):
                    llr[k] = np.dot(rmat, [top_x2[k], top_y2[k], 0])
                    site3 = np.dot(rmat, [site_x2, site_y2, 0])
                Li = np.min([np.max([a[0] for a in llr]), site3[0]])

                # -------------------------------------------------------------
                # Compute LD and save results into matrices
                # -------------------------------------------------------------

                Lrup = np.sqrt(Li * Li + self._Wrup * self._Wrup)
                LD[i, j] = np.log(Lrup) / np.log(Lrup_max)
                Ls[i, j] = Li
        self._LD = LD
        self._Ls = Ls

    def __computeDT(self, period):
        """
        Computes DT -- the distance taper term.
        """
        slat = self._lat
        slon = self._lon
        site_z = np.zeros_like(slat)
        ddict = get_distance("rrup", slat, slon, site_z, self._rup)
        Rrup = np.reshape(ddict["rrup"], (-1,))
        nsite = len(Rrup)

        if self._simpleDT:  # eqn 3.10
            R1 = 35
            R2 = 70
            DT = np.ones(nsite)
            ix = tuple([(Rrup > R1) & (Rrup < R2)])
            DT[ix] = 2 - Rrup[ix] / R1
            DT[Rrup >= R2] = 0
        else:  # eqn 3.9
            if period >= 1:
                R1 = 20 + 10 * np.log(period)
                R2 = 2 * (20 + 10 * np.log(period))
            else:
                R1 = 20
                R2 = 40
            DT = np.ones(nsite)
            # As written in report:
            # DT[(Rrup > R1) & (Rrup < R2)] = \
            #    2 - Rrup[(Rrup > R1) & (Rrup < R2)]/(20 + 10 * np.log(period))
            # Modification:
            DT[(Rrup > R1) & (Rrup < R2)] = 2 - Rrup[(Rrup > R1) & (Rrup < R2)] / R1
            # Note: it is not clear if the above modification is 'correct'
            #       but it gives results that make more sense
            DT[Rrup >= R2] = 0
        DT = np.reshape(DT, slat.shape)
        self._DT = DT

    def __getCenteringTerm(self):
        L = self._rup.getLength()
        CSSLP = -0.32 + 0.15 * np.log(L)
        CTRST = -0.20 + 0.09 * np.log(L)
        CNRML = -0.20 + 0.08 * np.log(L)
        if self._rake > 0:
            act = (CSSLP - CTRST) / 2
            cct = (CSSLP + CTRST) / 2
            self._xi_c = act * np.cos(2 * np.radians(self._rake)) + cct
        else:
            act = (CSSLP - CNRML) / 2
            cct = (CSSLP + CNRML) / 2
            self._xi_c = act * np.cos(2 * np.radians(self._rake)) + cct

    def __computeWP(self, period):
        """
        Computes WP -- the narrow-band multiplier.
        """
        # As written in report:
        # sig = 0.6
        # Tp = np.exp(1.27*self._M - 7.28)
        # Update (Rowshandel, personal communication)
        sig = 0.55
        Tp = np.exp(1.04 * self._M - 6.0)
        self._WP = np.exp(
            -(np.log10(period / Tp) * np.log10(period / Tp)) / (2 * sig * sig)
        )

    @classmethod
    def getPeriods(cls):
        """
        Returns a list of the periods for which the model is defined.
        """
        return cls.__periods


def _get_quad_slip_ds_ss(q, rake, cp, p):
    """
    Compute the DIP SLIP and STRIKE SLIP components of the unit slip vector in
    ECEF coords for a quad and rake angle.
    :param q:
        A quadrilateral.
    :param rake:
        Direction of motion of the hanging wall relative to the
        foot wall, as measured by the angle (deg) from the strike vector.
    :param cp:
        A 3x(n sub rupture) array giving center points of each sub rupture
        in ECEF coords.
    :param p:
        A 3x(n sub rupture) array giving the unit vectors of the propagation
        vector on each sub rupture in ECEF coords.
    :returns:
        List of dip slip and strike slip components (each is a matrix)
        of the unit slip vector in ECEF space.
    """
    # Get quad vertices, strike, dip
    P0, P1 = q[0:2]
    strike = P0.azimuth(P1)
    dip = utils.get_quad_dip(q)

    # Slip unit vectors in 'local' (i.e., z-up, x-east) coordinates
    d1_local = utils.get_local_unit_slip_vector_DS(strike, dip, rake)
    s1_local = utils.get_local_unit_slip_vector_SS(strike, dip, rake)

    # Convert to a column array
    d1_col = np.array([[d1_local.x], [d1_local.y], [d1_local.z]])
    s1_col = np.array([[s1_local.x], [s1_local.y], [s1_local.z]])

    # Define 'local' coordinate system
    qlats = [a.latitude for a in q]
    qlons = [a.longitude for a in q]
    proj = OrthographicProjection(
        np.min(qlons), np.max(qlons), np.min(qlats), np.max(qlats)
    )

    # Convert p and cp to geographic coords
    p0lat, p0lon, p0z = ecef2latlon(
        cp[0,],
        cp[1,],
        cp[2,],
    )
    p1lat, p1lon, p1z = ecef2latlon(
        cp[0,] + p[0,],
        cp[1,] + p[1,],
        cp[2,] + p[2,],
    )

    # Convert p to 'local' coords
    p0x, p0y = proj(p0lon, p0lat)
    p1x, p1y = proj(p1lon, p1lat)
    px = p1x - p0x
    py = p1y - p0y
    pz = p1z - p0z

    # Apply sign changes in 'local' coords
    s1mat = np.array(
        [
            [np.abs(s1_col[0]) * np.sign(px)],
            [np.abs(s1_col[1]) * np.sign(py)],
            [np.abs(s1_col[2]) * np.sign(pz)],
        ]
    )
    #                      [np.abs(s1_col[2])*np.ones_like(pz)]])

    dipsign = -np.sign(np.sin(np.radians(rake)))
    d1mat = np.array(
        [
            [d1_col[0] * np.ones_like(px) * dipsign],
            [d1_col[1] * np.ones_like(py) * dipsign],
            [d1_col[2] * np.ones_like(pz) * dipsign],
        ]
    )

    # Need to track 'origin'
    s0 = np.array([[0.0], [0.0], [0.0]])
    # Convert from 'local' to geographic coords
    s1_ll = proj(
        s1mat[0,][0],
        s1mat[1,][0],
        reverse=True,
    )
    d1_ll = proj(
        d1mat[0,][0],
        d1mat[1,][0],
        reverse=True,
    )
    s0_ll = proj(s0[0], s0[1], reverse=True)

    # And then back to ECEF:
    s1_ecef = latlon2ecef(
        s1_ll[1][0],
        s1_ll[0][0],
        s1mat[2,][0],
    )
    d1_ecef = latlon2ecef(
        d1_ll[1][0],
        d1_ll[0][0],
        d1mat[2,][0],
    )
    s0_ecef = latlon2ecef(s0_ll[1], s0_ll[0], s0[2])
    s00 = s0_ecef[0].reshape(-1)
    s01 = s0_ecef[1].reshape(-1)
    s02 = s0_ecef[2].reshape(-1)
    d_mat = np.array(
        [
            d1_ecef[0].reshape(-1) - s00,
            d1_ecef[1].reshape(-1) - s01,
            d1_ecef[2].reshape(-1) - s02,
        ]
    )
    s_mat = np.array(
        [
            s1_ecef[0].reshape(-1) - s00,
            s1_ecef[1].reshape(-1) - s01,
            s1_ecef[2].reshape(-1) - s02,
        ]
    )
    return d_mat, s_mat


def _rotation_matrix(axis, theta):
    """
    Return the rotation matrix associated with counterclockwise rotation about
    the given axis by theta radians.
    Source: Response by 'unutbu' in this thread:
    http://stackoverflow.com/questions/6802577/python-rotation-of-3d-vector
    """
    axis = np.asarray(axis)
    theta = np.asarray(theta)
    axis = axis / np.sqrt(np.dot(axis, axis))
    a = np.cos(theta / 2)
    b, c, d = -axis * np.sin(theta / 2)
    aa, bb, cc, dd = a * a, b * b, c * c, d * d
    bc, ad, ac, ab, bd, cd = b * c, a * d, a * c, a * b, b * d, c * d
    return np.array(
        [
            [aa + bb - cc - dd, 2 * (bc + ad), 2 * (bd - ac)],
            [2 * (bc - ad), aa + cc - bb - dd, 2 * (cd + ab)],
            [2 * (bd + ac), 2 * (cd - ab), aa + dd - bb - cc],
        ]
    )



import numpy as np
from scipy.interpolate import RegularGridInterpolator
import itertools as it

from esi_shakelib.correlation.ccf_base import CrossCorrelationBase
from esi_core.shakemap.covariance_matrix import eval_lb_correlation

# Periods that apply to the axes in the cross-correlation tables
Tlist = np.array([0.01, 0.1, 0.2, 0.5, 1, 2, 5, 7.5, 10.0001])

# Table II. Short range coregionalization matrix, B1
B1 = np.array(
    [
        [0.29, 0.25, 0.23, 0.23, 0.18, 0.10, 0.06, 0.06, 0.06],
        [0.25, 0.30, 0.20, 0.16, 0.10, 0.04, 0.03, 0.04, 0.05],
        [0.23, 0.20, 0.27, 0.18, 0.10, 0.03, 0.00, 0.01, 0.02],
        [0.23, 0.16, 0.18, 0.31, 0.22, 0.14, 0.08, 0.07, 0.07],
        [0.18, 0.10, 0.10, 0.22, 0.33, 0.24, 0.16, 0.13, 0.12],
        [0.10, 0.04, 0.03, 0.14, 0.24, 0.33, 0.26, 0.21, 0.19],
        [0.06, 0.03, 0.00, 0.08, 0.16, 0.26, 0.37, 0.30, 0.26],
        [0.06, 0.04, 0.01, 0.07, 0.13, 0.21, 0.30, 0.28, 0.24],
        [0.06, 0.05, 0.02, 0.07, 0.12, 0.19, 0.26, 0.24, 0.23],
    ]
)

# Table III. Long range coregionalization matrix, B2
B2 = np.array(
    [
        [0.47, 0.40, 0.43, 0.35, 0.27, 0.15, 0.13, 0.09, 0.12],
        [0.40, 0.42, 0.37, 0.25, 0.15, 0.03, 0.04, 0.00, 0.03],
        [0.43, 0.37, 0.45, 0.36, 0.26, 0.15, 0.09, 0.05, 0.08],
        [0.35, 0.25, 0.36, 0.42, 0.37, 0.29, 0.20, 0.16, 0.16],
        [0.27, 0.15, 0.26, 0.37, 0.48, 0.41, 0.26, 0.21, 0.21],
        [0.15, 0.03, 0.15, 0.29, 0.41, 0.55, 0.37, 0.33, 0.32],
        [0.13, 0.04, 0.09, 0.20, 0.26, 0.37, 0.51, 0.49, 0.49],
        [0.09, 0.00, 0.05, 0.16, 0.21, 0.33, 0.49, 0.62, 0.60],
        [0.12, 0.03, 0.08, 0.16, 0.21, 0.32, 0.49, 0.60, 0.68],
    ]
)

# Table IV. Nugget effect coregionalization matrix, B3
B3 = np.array(
    [
        [0.24, 0.22, 0.21, 0.09, -0.02, 0.01, 0.03, 0.02, 0.01],
        [0.22, 0.28, 0.20, 0.04, -0.05, 0.00, 0.01, 0.01, -0.01],
        [0.21, 0.20, 0.28, 0.05, -0.06, 0.00, 0.04, 0.03, 0.01],
        [0.09, 0.04, 0.05, 0.26, 0.14, 0.05, 0.05, 0.05, 0.04],
        [-0.02, -0.05, -0.06, 0.14, 0.20, 0.07, 0.05, 0.05, 0.05],
        [0.01, 0.00, 0.00, 0.05, 0.07, 0.12, 0.08, 0.07, 0.06],
        [0.03, 0.01, 0.04, 0.05, 0.05, 0.08, 0.12, 0.10, 0.08],
        [0.02, 0.01, 0.03, 0.050, 0.05, 0.07, 0.10, 0.10, 0.09],
        [0.01, -0.01, 0.01, 0.04, 0.05, 0.06, 0.08, 0.09, 0.09],
    ]
)


class LothBaker2013(CrossCorrelationBase):
    """
    Created by Christophe Loth, 12/18/2012
    Pythonized and vectorized by C. Bruce Worden, 3/15/2017
    Updated with the erratum tables by C. Bruce Worden, 1/13/2021
    Compute the spatial correlation of epsilons for the NGA ground motion
    models

    The function is strictly empirical, fitted over the range the range
    0.01s <= t1, t2 <= 10s

    Documentation is provided in the following paper:
    Loth, C., and Baker, J. W. (2013). “A spatial cross-correlation model of
    ground motion spectral accelerations at multiple periods.”
    Earthquake Engineering & Structural Dynamics, 42, 397-417.

    Updated to include the erratum of:
    Loth, C., and Baker, J. W. (2019). "Erratum: A spatial cross-correlation
    model for ground motion spectral accelerations at multiple periods."
    Earthquake Engineering & Structural Dynamics, 49(3), 315-316.
    https://doi.org/10.1002/eqe.3233
    """

    def __init__(self, periods):
        """
        Create an instance of LB13.

        Args:
            periods (numpy.array): An array of periods that will be requested
                from the function. Values must be [0.01 -> 10.0], and must be
                sorted from smallest to largest.

        Returns:
            An instance of :class:`LothBaker2013`.
        """

        if np.any(periods < 0.01):
            raise ValueError("The periods must be greater or equal to 0.01s")
        if np.any(periods > 10):
            raise ValueError("The periods must be less or equal to 10s")

        nper = np.size(periods)

        rgi1 = RegularGridInterpolator((Tlist, Tlist), B1, method="linear")
        rgi2 = RegularGridInterpolator((Tlist, Tlist), B2, method="linear")
        rgi3 = RegularGridInterpolator((Tlist, Tlist), B3, method="linear")

        #
        # Build new tables with entries at the periods we will use
        #
        tlist = list(it.product(periods, periods))
        self.b1 = rgi1(tlist).reshape((nper, nper))
        self.b2 = rgi2(tlist).reshape((nper, nper))
        self.b3 = rgi3(tlist).reshape((nper, nper))

        db1 = np.diagonal(B1)
        db2 = np.diagonal(B2)
        db3 = np.diagonal(B3)

        rgid1 = RegularGridInterpolator((Tlist,), db1, method="linear")
        rgid2 = RegularGridInterpolator((Tlist,), db2, method="linear")
        rgid3 = RegularGridInterpolator((Tlist,), db3, method="linear")

        id1 = rgid1(periods)
        id2 = rgid2(periods)
        id3 = rgid3(periods)

        np.fill_diagonal(self.b1, id1)
        np.fill_diagonal(self.b2, id2)
        np.fill_diagonal(self.b3, id3)

    def getCorrelation(self, ix1, ix2, h):
        """
        Compute the correlation between two periods and a separation distance
        of h.

        The indices (ix1 and ix2) and h must have the same dimensions. The
        indices may be equal, and there is no restriction on which one is
        larger. The indices refer to periods in the 'period' argument to the
        class constructor. The result is stored in h.

        Args:
            ix1 (2D, C-contiguous numpy array)):
                The indices of the first period of interest.
            ix2 (2D, C-contiguous numpy array)):
                The indices of the second period of interest.
            h (2D, C-contiguous numpy array)):
                The separation distance between two sites (units of km).
                h will be returned with the result, so it must be
                copied if the values in h are to be preserved.

        Returns:
            h (numpy array):
                The predicted correlation coefficient. The output array
                will have the same shape as the inputs.

        """
        # Verify the validity of input arguments
        if np.any(h < 0):
            raise ValueError("The separation distance must be positive")
        if np.shape(ix1) != np.shape(ix2) != np.shape(h):
            raise ValueError("The input arguments must all have the same dimensions")

        #
        # Index into the arrays to get the coefficients corresponding to the
        # periods of interest.
        #
        # These variables are used in ne.evaluate but unseen by linter
        # b1 = self.b1[ix1, ix2]  # noqa
        # b2 = self.b2[ix1, ix2]  # noqa
        # b3 = self.b3[ix1, ix2]  # noqa
        # afact = -3.0 / 20.0  # noqa
        # bfact = -3.0 / 70.0  # noqa
        #
        # Compute the correlation coefficient (Equation 42)
        #
        # rho = ne.evaluate(
        #     "b1 * exp(h * afact) + b2 * exp(h * bfact) + (h == 0) * b3")
        rho = eval_lb_correlation(self.b1, self.b2, self.b3, ix1, ix2, h)

        return rho



#!/usr/bin/env python

# Standard library imports
import os.path
import sys

# Third party imports
import numpy as np
import pytest

# Local imports
from esi_shakelib.conversions.imt.newmark_hall_1982 import NewmarkHall1982


homedir = os.path.dirname(os.path.abspath(__file__))  # where is this script?
shakedir = os.path.abspath(os.path.join(homedir, "..", "..", ".."))
sys.path.insert(0, shakedir)


def test_newmarkhall1982():
    # Inputs
    PSA10in = np.log(0.1)
    sd = 0.6
    nh82 = NewmarkHall1982()

    # Test that the correct inputs are returned for valid outputs
    input1 = nh82.getInputIMT("pgv  ")
    input2 = nh82.getInputIMT("PGV")
    assert len(input1) == len(input2) == 1
    assert input1[0] == input2[0] == "PSA10"

    # Test that invalid outputs return None
    input3 = nh82.getInputIMT("INVALID")
    assert input3 == None

    # Test valid conversions
    PGVout = nh82.convertAmps("psa10", "PGV", PSA10in)
    PGVsdout = nh82.convertSigmas("psa10", "PGV", sd)
    mfact = nh82.getConversionFactor()
    lnsig = nh82.getLnSigma()
    assert abs(PGVout - np.log(9.46658)) < 0.001
    assert abs(PGVsdout - 0.790489) < 0.001
    assert abs(mfact - 94.6658) < 0.001
    assert abs(lnsig - 0.5146578) < 0.001

    # Test invalid conversions
    with pytest.raises(Exception) as a:
        tmp = nh82.convertAmps("INVALID", "PSA05", PGVin)
    with pytest.raises(Exception) as a:
        tmp = nh82.convertAmps("PGV", "INVALID", PGVin)
    with pytest.raises(Exception) as a:
        tmp = nh82.convertAmps("INVALID", "INVALID", PGVin)


if __name__ == "__main__":
    test_newmarkhall1982()



import numpy as np

from openquake.hazardlib import const
from openquake.hazardlib.contexts import ContextMaker


def gmpe_gmas(gmpe, cx, imt, stddev_types):
    """ """
    if isinstance(imt, list):
        nlist = len(imt)
        imtstr = imt[0].string
        imtl = imt
    else:
        nlist = 1
        imtstr = imt.string
        imtl = [imt]
    if gmpe.compute.__annotations__.get("ctx") is np.recarray:
        magstr = f"{cx.mag[0]:.2f}"
        param = dict(
            imtls={imtstr: [0]},
            maximum_distance=4000,
            truncation_level=3,
            investigation_time=1.0,
            mags=[magstr],
        )
        cmaker = ContextMaker("*", [gmpe], param)
        if not isinstance(cx, np.ndarray):
            cx = cmaker.recarray([cx])

    N = len(cx)
    mean = np.zeros((nlist, N))
    sig = np.zeros((nlist, N))
    tau = np.zeros((nlist, N))
    phi = np.zeros((nlist, N))
    try:
        gmpe.compute(cx, imtl, mean, sig, tau, phi)
    except NotImplementedError:
        mean, stddevs = gmpe.get_mean_and_stddevs(
            cx, cx, cx, imt, stddev_types
        )
        return mean, stddevs
    except Exception as exc:
        raise exc
    stddevs = []
    for i in range(nlist):
        for stddev_type in stddev_types:
            if stddev_type == const.StdDev.TOTAL:
                stddevs.append(sig[i])
            elif stddev_type == const.StdDev.INTER_EVENT:
                stddevs.append(tau[i])
            elif stddev_type == const.StdDev.INTRA_EVENT:
                stddevs.append(phi[i])
    return mean, stddevs



"""
Module to find coefficient tables and SA params of GMPEs.
"""


def get_gmpe_coef_table(gmpe):
    """
    Method for finding the (or "a") GMPE table.

    Notes:

      *  The reason for the complexity here is that there can be multiple
         coefficient tables, and some of them may not have the sa_coeffs
         attribute, which is the main reason for getting the table.
      *  We are also assuming that if there are more than one  coefficient
         table, the range of periods will be the same across all of the
         tables.

    Args:
        gmpe (GMPE): An OQ GMPE instance.

    Returns:
        The associated coefficient table.

    """
    stuff = dir(gmpe)
    coef_list = [s for s in stuff if "COEFFS" in s]
    for coef_sel in coef_list:
        cobj = getattr(gmpe, coef_sel)
        if "sa_coeffs" in dir(cobj):
            return cobj
    raise NotImplementedError(f"GMPE {gmpe} does not contain sa_coeffs attribute.")


def get_gmpe_sa_periods(gmpe):
    """
    Method to extract the SA periods defined by a GMPE.

    Args:
        gmpe (GMPE): A GMPE instance.

    Retunrs:
        list: List of periods.

    """
    if gmpe == "[NGAEast]":
        per = gmpe.per_array
    else:
        ctab = get_gmpe_coef_table(gmpe).sa_coeffs
        ilist = list(ctab.keys())
        per = [i.period for i in ilist]
    return per



# Standard library imports
from abc import ABC, abstractmethod

from openquake.hazardlib.const import IMC


class ComponentConverter(ABC):
    """Base class for implementing conversions between components."""

    def checkUnknown(self):
        """
        Checks if imc_in or imc_out is an unknown type.

        If of an unknown type, it is automatically set
        to GEOMETRIC_MEAN.
        """
        if self.imc_in not in self.conversion_graph:
            self.imc_in = IMC.GEOMETRIC_MEAN
        if self.imc_out not in self.conversion_graph:
            self.imc_out = IMC.GEOMETRIC_MEAN

    def convertAmps(self, imt, amps, rrups=None, mag=None):
        """
        Return an array of amps converted from one IMC to another.

        Note: This can be used to perform chained conversions.

        Args:
            imt (OpenQuake IMT): The intensity measure type of the input
                ground motions. Valid IMTs are PGA, PGV, and SA.
            amps (array): A numpy array of the (logged) ground motions
                to be converted.
            rrups (array): A numpy array of the same shape as amps,
                containing the rupture distances of the ground motions.
                Default is None.
            mag (float): The earthquake magnitude. Default is None.

        Returns:
            array: A numpy array of converted ground motions.
        """
        for idx, path in enumerate(self.path):
            if idx != len(self.path) - 1:
                self.imc_in = path
                self.imc_out = self.path[idx + 1]
                amps = self.convertAmpsOnce(imt, amps, rrups, mag)
        return amps

    @abstractmethod
    def convertAmpsOnce(self, imt, amps, rrups=None, mag=None):
        """
        Return an array of amps converted from one IMC to another.

        Note: This does not implement chained conversions.

        Args:
            imt (OpenQuake IMT): The intensity measure type of the input
                ground motions. Valid IMTs are PGA, PGV, and SA.
            amps (array): A numpy array of the (logged) ground motions
                to be converted.
            rrups (array): A numpy array of the same shape as amps,
                containing the rupture distances of the ground motions.
            mag (float): The earthquake magnitude.

        Returns:
            array: A numpy array of converted ground motions.
        """
        pass

    def convertSigmas(self, imt, sigmas):
        """
        Return an array of standard deviations converted from one IMC
        to another.

        Note:  This can be used to perform chained conversions.

        Args:
            imt (OpenQuake IMT): The intensity measure type of the input
                ground motions. Valid IMTs are PGA, PGV, and SA.
            sigmas (array): A numpy array of the standard deviations of
                the logged ground motions.

        Returns:
            array: A numpy array of converted standard deviations.
        """
        for idx, path in enumerate(self.path):
            if idx != len(self.path) - 1:
                self.imc_in = path
                self.imc_out = self.path[idx + 1]
                sigmas = self.convertSigmasOnce(imt, sigmas)
        return sigmas

    @abstractmethod
    def convertSigmasOnce(self, imt, sigmas):
        """
        Return an array of standard deviations converted from one IMC
        to another.

        Note: This does not implement chained conversions.

        Args:
            imt (OpenQuake IMT): The intensity measure type of the input
                ground motions. Valid IMTs are PGA, PGV, and SA.
            sigmas (array): A numpy array of the standard deviations of
                the logged ground motions.

        Returns:
            array: A numpy array of converted standard deviations.
        """
        pass

    @staticmethod
    def pathSearch(graph, imc_in, imc_out):
        """
        Helper method to create a "path" to convert one IMC to another.

        This can be used for chain conversions

        Args:
            graph (Dictionary): Dictionary of sets describing possible
                conversions between IMCs.
            imc_in (IMC): OpenQuake IMC type of the input amp array.
            imc_out (IMC): Desired OpenQuake IMC type of the output amps.

        Returns:
            list: IMCs as a path for to convert one IMC to another.
        """
        queue = [(imc_in, [imc_in])]
        while queue:
            (key, path) = queue.pop(0)
            for next_path in graph[key] - set(path):
                if next_path == imc_out:
                    yield path + [next_path]
                else:
                    queue.append((next_path, path + [next_path]))

    def getShortestPath(self, graph, imc_in, imc_out):
        """
        Create a "path" to convert one IMC to another.

        This can be used for chain conversions

        Args:
            graph (Dictionary): Dictionary of sets describing possible
                conversions between IMCs.
            imc_in (IMC): OpenQuake IMC type of the input amp array.
            imc_out (IMC): Desired OpenQuake IMC type of the output amps.

        Returns:
            list: IMCs as a path for to convert one IMC to another.

        Raises:
            ValueError if no path is found.
        """
        try:
            return next(self.pathSearch(graph, imc_in, imc_out))
        except StopIteration:
            if imc_in == imc_out:
                return [imc_in, imc_out]
            else:
                raise ValueError(
                    f"No possible conversion between {imc_in!r} and {imc_out!r}."
                )

    @abstractmethod
    def _verifyConversion(self, imc_in, imc_out=None):
        """
        Helper method to ensure that the conversion is possible.

        Args:
            imc_in (IMC): OpenQuake IMC type of the input amp array.
            imc_out (IMC): Desired OpenQuake IMC type of the output amps.
                Default is None.

        Raises:
            ValueError if imc_in or imc_out are not valid..
        """
        pass

    def imc_from_str(self, imcstring):
        """
        Convert a string to one of the OQ IMC types. Return None if the string
        does not match any of the strings in the Enum.

        Args:
            imcstring (str): A string corresponding to one of the OQ IMC types.

        Returns:
            IMC: One of the OQ IMC types. Returns None if nothing matches.
        """
        for imc in IMC:
            if imc.value.lower() == imcstring.lower():
                return imc
        return None



"""
Module to simplify importing the OQ implementation of the
NGA-East GMPE suite.
"""

import os
import re
import sys
import logging
import copy

import pandas as pd
import numpy as np
from scipy.interpolate import RectBivariateSpline

# This import should be changed when the code is advanced beyond python 3.9 to:
# import importlib.resources as importlib_resources
import importlib_resources

from openquake.hazardlib.gsim.base import GMPE
from openquake.hazardlib import const
from openquake.hazardlib import imt as IMT
from openquake.hazardlib.gsim.usgs_ceus_2019 import NGAEastUSGSGMPE
from openquake.hazardlib.gsim.gmpe_table import _return_tables

from esi_shakelib.multiutils import gmpe_gmas
from esi_shakelib.conversions.imt.abrahamson_bhasin_2020 import AbrahamsonBhasin2020

# Max distance for evaluating NGAEast. This *should* be 1500, but due to what
# appears to be a floating point precision issue, we get a division by zero
# error at a value of 1500 returning nans. So we have to cap the distance at a
# value slightly smaller.

# This is also used to zero out the results at greater distances.

MAX_RRUP = 1499.9


class NGAEast(GMPE):
    """
    Returns NGA East GMPE that combines all of the individual NGAEastUSGSGMPE
    GMPEs.
    """

    DEFINED_FOR_TECTONIC_REGION_TYPE = const.TRT.STABLE_CONTINENTAL
    DEFINED_FOR_INTENSITY_MEASURE_COMPONENT = const.IMC.RotD50
    DEFINED_FOR_INTENSITY_MEASURE_TYPES = set([IMT.PGA, IMT.PGV, IMT.SA])

    DEFINED_FOR_STANDARD_DEVIATION_TYPES = set(
        [const.StdDev.TOTAL, const.StdDev.INTER_EVENT, const.StdDev.INTRA_EVENT]
    )
    REQUIRES_SITES_PARAMETERS = set(("vs30",))
    REQUIRES_RUPTURE_PARAMETERS = set(("mag",))
    REQUIRES_DISTANCES = set(("rrup",))

    TABLE_PATHS = os.listdir(NGAEastUSGSGMPE.PATH)

    NGA_EAST_USGS_WEIGHT = 0.667
    NGA_EAST_SEED_WEIGHT = 0.333

    # Sigma models and their weights
    SIGMA_MODS = ["EPRI", "PANEL"]
    SIGMA_WEIGHTS = [0.8, 0.2]

    # For small magnitude extrapolation
    __path__ = importlib_resources.files("esi_shakelib") / "gmpe" / "nga_east_small_mag"
    __efile__ = __path__ / "nga-east-smallM_slopes.txt"
    with importlib_resources.as_file(__efile__) as path:
        SMALL_M_SLOPE = np.loadtxt(path)
    __efile__ = __path__ / "nga-east-smallM_slope_pga.txt"
    with importlib_resources.as_file(__efile__) as path:
        SMALL_M_SLOPE_PGA = np.loadtxt(path)
    __efile__ = __path__ / "nga-east-smallM_slope_pgv.txt"
    with importlib_resources.as_file(__efile__) as path:
        SMALL_M_SLOPE_PGV = np.loadtxt(path)
    __efile__ = __path__ / "nga-east-smallM_slope_distances.txt"
    with importlib_resources.as_file(__efile__) as path:
        SMALL_M_DIST = np.loadtxt(path)
    __efile__ = __path__ / "nga-east-smallM_slope_periods.txt"
    with importlib_resources.as_file(__efile__) as path:
        SMALL_M_PER = np.loadtxt(path)

    # -------------------------------------------------------------------------
    # To simplify, use the COLLAPSED branch, but cannot get inter and intra
    # event standard deviations in this case.
    # SIGMA_MODS = ["COLLAPSED"]
    # SIGMA_WEIGHTS = [1.0]

    def __init__(self):
        gmpes = []
        sigma_wts = []
        all_table_paths = []
        for i, sigma_mod in enumerate(self.SIGMA_MODS):
            for table_path in self.TABLE_PATHS:
                gmpe = NGAEastUSGSGMPE(gmpe_table=table_path, sigma_model=sigma_mod)
                gmpes.append(gmpe)
                sigma_wts.append(self.SIGMA_WEIGHTS[i])
                all_table_paths.append(table_path)
        self.gmpes = gmpes
        self.sigma_weights = np.array(sigma_wts)
        self.ALL_TABLE_PATHS = all_table_paths

        mod = importlib_resources.files("esi_shakelib") / "gmpe" / "nga_east_tables"
        efile = mod / "nga-east-usgs-weights.dat"
        with importlib_resources.as_file(efile) as path:
            self.NGA_EAST_USGS = pd.read_csv(path)
        efile = mod / "nga-east-seed-weights.dat"
        with importlib_resources.as_file(efile) as path:
            self.NGA_EAST_SEEDS = pd.read_csv(path)

        # Parse the periods of the columns
        self.per_idx_start = 1
        self.per_idx_end = -2
        per_list_str = self.NGA_EAST_USGS.keys().tolist()[
            self.per_idx_start : self.per_idx_end
        ]
        self.per_array = np.array(
            [float(p.replace("SA", "").replace("P", ".")) for p in per_list_str]
        )

    def get_mean_and_stddevs(self, sites, rx, dists, imt, stddev_types):
        if not isinstance(imt, list):
            imt = [imt]
        nstd = len(stddev_types)
        mean = []
        std = []
        for iimt in imt:
            mm, ss = self.__get_mean_and_stddevs__(sites, rx, dists, iimt, stddev_types)
            mean.append(mm)
            for i, sdt in enumerate(stddev_types):
                std.append(ss[i])
        return mean, std

    def __get_mean_and_stddevs__(self, sites, rx, dists, imt, stddev_types):
        # List of GMPE weights, which is the product of the the branch weights
        # for the seed models vs the NGA East resampled models as well as the
        # weights for the indivudual GMPES as defined by Petersen et al. (2019)
        #
        # Note that the NGA East resampled models are a function of spectral
        # period.
        #
        # NGA East Seeds (1/3)
        #     ├── B_bca10d (0.06633), wts = 0.333 * 0.06633 = 0.02208789
        #     ├── B_ab95 (0.02211), wts = 0.333 * 0.02211 = 0.00736263
        #     ...
        # NGA East Resampled or "USGS" (2/3)
        #     ├── Model 1 (0.1009 for PGA), wts = 0.667 * 0.1009 = 0.0673003
        #     ├── Model 2 (0.1606 for PGA), wts = 0.667 * 0.1606 = 0.1071202
        #     ...
        #
        wts = [0] * len(self.gmpes)

        # Is IMT PGA or PGV?
        is_pga = imt == IMT.PGA()
        is_pgv = imt == IMT.PGV()

        # Is magnitude less than 4? If so, we will need to set it to 4.0 and
        # then extrapolate the tables at the end.
        # But... in the brave new world of new OpenQuake, sites, rx, and
        # dists are all exactly the same object, so when we copy rx to
        # rup, and change it, as well as when we change the distances
        # below, we need to do it all in rup, then pass rup as all three
        # contexts when we call the gmpe.get_mean_and_stddevs()
        rup = copy.deepcopy(rx)
        if np.any(rup.mag < 4.0):
            is_small_mag = True
            delta_mag = rup.mag - 4.0
            rup.mag = 4.0
        else:
            is_small_mag = False

        for i, tp in enumerate(self.ALL_TABLE_PATHS):
            if "usgs" in tp:
                # Get model number from i-th path using regex
                mod_num = int(re.search(r"\d+", tp).group())
                coefs = np.array(self.NGA_EAST_USGS.iloc[mod_num - 1])
                # Is the IMT PGA, PGA, or SA?
                if is_pga:
                    iweight = coefs[-2]
                elif is_pgv:
                    iweight = coefs[-1]
                else:
                    # For SA, need to interpolate; we'll use log-period and
                    # linear-weight interpolation.
                    iweight = np.interp(
                        np.log(imt.period),
                        np.log(self.per_array),
                        coefs[self.per_idx_start : self.per_idx_end],
                    )
                wts[i] = self.NGA_EAST_USGS_WEIGHT * iweight
            else:
                # Strip off the cruft to get the string we need to match
                str_match = tp.replace("nga_east_", "").replace(".hdf5", "")
                matched = self.NGA_EAST_SEEDS[self.NGA_EAST_SEEDS["model"] == str_match]
                if len(matched):
                    iweight = self.NGA_EAST_SEEDS[
                        self.NGA_EAST_SEEDS["model"] == str_match
                    ].iloc[0, 1]
                    wts[i] = self.NGA_EAST_SEED_WEIGHT * iweight

        total_gmpe_weights = self.sigma_weights * wts

        if not np.allclose(np.sum(total_gmpe_weights), 1.0):
            raise ValueError("Weights must sum to 1.0.")

        mean = np.full_like(sites.vs30, 0)
        stddevs = []
        for i in range(len(stddev_types)):
            stddevs.append(np.full_like(sites.vs30, 0))

        # Apply max distance to dists.rrup -->> now rup.rrup
        np.clip(rup.rrup, 0, MAX_RRUP)

        #
        # Some models don't have PGV terms, so we will make PSA for them
        # and then use the conditional conversion to get PGV
        #
        if is_pgv:
            ab2020 = AbrahamsonBhasin2020(rup.mag)
            vimt = IMT.SA(ab2020.getTref())

        # Loop over gmpes
        for i, gm in enumerate(self.gmpes):
            if is_pgv:
                # Is PGV and also not available for gm?
                try:
                    _ = _return_tables(gm, rup.mag, imt, "IMLs")
                except KeyError:
                    #
                    # No table for PGV, compute vimt, then convert to PGV
                    #
                    vmean, vstddevs = gmpe_gmas(gm, rup, vimt, stddev_types)
                    tmean, tstddevs = ab2020.getPGVandSTDDEVS(
                        vmean, vstddevs, stddev_types, rup.rrup, rup.vs30
                    )
                except Exception:
                    logging.error(f"Unexpected error: {sys.exc_info()[0]}")
                else:
                    #
                    # Table exists for PGV, proceed normally
                    #
                    tmean, tstddevs = gmpe_gmas(gm, rup, imt, stddev_types)
            else:
                tmean, tstddevs = gmpe_gmas(gm, rup, imt, stddev_types)

            mean += tmean[0] * total_gmpe_weights[i]
            for j, sd in enumerate(tstddevs):
                stddevs[j] += sd * total_gmpe_weights[i]

        # Zero out values at distances beyond the range for which NGA East
        # was defined. -->> was dists.rrup, now rup.rrup
        mean[rup.rrup > MAX_RRUP] = -999.0

        # Do we need to extrapolate for small magnitude factor?
        if is_small_mag:
            if is_pga:
                slopes = np.interp(
                    np.log(rup.rrup), np.log(self.SMALL_M_DIST), self.SMALL_M_SLOPE_PGA
                )
            elif is_pgv:
                slopes = np.interp(
                    np.log(rup.rrup), np.log(self.SMALL_M_DIST), self.SMALL_M_SLOPE_PGV
                )
            else:
                interp_obj = RectBivariateSpline(
                    np.log(self.SMALL_M_DIST),
                    np.log(self.SMALL_M_PER),
                    self.SMALL_M_SLOPE,
                    kx=1,
                    ky=1,
                )
                slopes = interp_obj.ev(np.log(rup.rrup), np.log(imt.period))
            mean = mean + slopes * delta_mag

        return mean, stddevs



# Standard library imports
import copy

# Third party imports
import numpy as np

# Local imports
from esi_shakelib.conversions.convert_imt import IMTConverter


class NewmarkHall1982(IMTConverter):
    """
    Class for conversion between PGA and PSA10 by Newmark and Hall (1982).

    - PSA10 stands for spectral acceleration with oscillator period of 1.0 sec
    - PGV is peak ground velocity.

    The conversion factor is 37.27*2.54.
    Note that 2.54 is the conversion factor to convert from cm/s to in/s and
    37.27 is derived from SA(f) = 1.65*(2*pi*V*f)/386.09, where:

    - SA(f) is spectral acceleration at frequency f (in g)
    - f is the frequency of interest
    - V is the velocity
    - 1.65 is the N&H amplification factor for velocity at 5% damping
    - 386.09 is the acceleration of gravity in inches per second per g

    The sigma factor was computed from an average sigma value determined
    by plotting PGV/PSA10 versus Distance for earthquakes with magnitudes
    greater than or equal to 5.0.

    References:
        Newmark, N. M., & Hall, W. J. (1982). Earthquake spectra and design.
        Earthquake Engineering Research Institute, El Cerrito, California.
    """

    def __init__(self):
        super().__init__()
        # output_input dictionary where the key is the output
        # and the value is a list of the possible inputs
        self.output_input = {"PGV": ["PSA10"]}
        self._lnfact = np.log(37.27 * 2.54)
        self.conversion_factor = np.exp(self._lnfact)
        self._lnsigma = 0.5146578

    def convertAmps(self, imt_in, imt_out, imt):
        """
        Returns an array of converted IMT amplitude values.

        Args:
            imt_in (str): OQ intensity measure type. Same as type as the input
                values defined by the imt variable.
            imt_out (str): OQ intensity measure type that the values will
                be converted to.
            imt (OpenQuake IMT): The intensity measurements of the input
                ground motions. Valid IMTs are PGV, and SA.

        Returns:
            array: Numpy array of amps converted from imt_in to imt_out.

        Raises:
            ValueError: If not a valid conversion.
        """
        self._verifyConversion(imt_in, imt_out)
        imt_in = imt_in.upper().strip()
        imt_out = imt_out.upper().strip()
        conversion_factor = self._lnfact
        if imt_in == "PSA10" and imt_out == "PGV":
            new_imt = self._convertToPGV(imt, conversion_factor)
        else:
            raise ValueError(f"No conversion available from {imt_in!r} to {imt_out!r}")
        return new_imt

    def convertSigmas(self, imt_in, imt_out, sigma):
        """
        Convert standard deviation to natural log units.

        Args:
            imt_in (str): OQ intensity measure type. Same as type as the input
                values defined by the imt variable.
            imt_out (str): OQ intensity measure type that the values will
                be converted to.
            sigma (array): Numpy array or float of standard deviation of PGV from a GMPE;
                units must be natural log.

        Returns:
            array: Converted standard deviations with natural log units.
        """
        self._verifyConversion(imt_in, imt_out)
        lnsigma = self._lnsigma
        sigmaTot = np.sqrt((sigma ** 2) + (lnsigma ** 2))
        return sigmaTot

    def getLnSigma(self):
        """
        Returns:
            float: The the estimate of the logarithmic standard deviation of the PGV
                predicted by Newmark and Hall (1982).
        """
        return copy.copy(self._lnsigma)

    @staticmethod
    def _convertToPGV(psa10, lnfact):
        """
        Convert PSA10 (spectral acceleration with oscillator period of 1.0 sec)
        to PGV.

        **Important:** PSA10 and sigma must be logarithmic units.

        Args:
            psa10 (array): OQ intensity measure type. Same as type as the input
                values defined by the imt variable.
            lnfact (float): Conversion factor.

        Returns:
            array: Converted PGV values with natural log of cm/s units.
        """
        pgv = psa10 + lnfact
        return pgv



"""
Module to implement CB14 basin and M9 modifications to Zhao et al. (2006) GSIM.
"""

from esi_shakelib.gmpe.modifiable_gmpes.m9_basin import M9BasinTerm


class ZhaoEtAl2006SInterBasin(M9BasinTerm):
    """
    Returns basin modification of the Zhao et al. (2009) interface GSIM.
    """

    m9 = False
    SJ = 0

    def __init__(self):
        super().__init__(gmpe_name="ZhaoEtAl2006SInter")


class ZhaoEtAl2006SInterM9Basin(M9BasinTerm):
    """
    Returns M9+basin modification of the Zhao et al. (2009) interface GSIM.
    """

    m9 = True
    SJ = 0

    def __init__(self):
        super().__init__(gmpe_name="ZhaoEtAl2006SInter")


class ZhaoEtAl2006SSlabBasin(M9BasinTerm):
    """
    Returns basin modification of the Zhao et al. (2009) slab GSIM.
    """

    m9 = False
    SJ = 0

    def __init__(self):
        super().__init__(gmpe_name="ZhaoEtAl2006SSlab")


class ZhaoEtAl2006SInterJapanM9Basin(M9BasinTerm):
    """
    Returns Japan M9+basin modification of the Zhao et al. (2009) interface GSIM.
    """

    m9 = True
    SJ = 1

    def __init__(self):
        super().__init__(gmpe_name="ZhaoEtAl2006SInter")


class ZhaoEtAl2006SSlabJapanBasin(M9BasinTerm):
    """
    Returns Japan basin modification of the Zhao et al. (2009) slab GSIM.
    """

    m9 = False
    SJ = 1

    def __init__(self):
        super().__init__(gmpe_name="ZhaoEtAl2006SSlab")


class ZhaoEtAl2006SInterCascadiaM9Basin(M9BasinTerm):
    """
    Returns Cascadia M9+basin modification of the Zhao et al. (2009) interface GSIM.
    """

    m9 = True
    SJ = 0

    def __init__(self):
        super().__init__(gmpe_name="ZhaoEtAl2006SInterCascadia")


class ZhaoEtAl2006SSlabCascadiaBasin(M9BasinTerm):
    """
    Returns Cascadia basin modification of the Zhao et al. (2009) slab GSIM.
    """

    m9 = False
    SJ = 0

    def __init__(self):
        super().__init__(gmpe_name="ZhaoEtAl2006SSlabCascadia")



#!/usr/bin/env python

# stdlib imports
import time
import datetime
import io
import os.path
import random
import string
import sys
import tempfile

# third party imports
import numpy as np
import pytest
from esi_utils_io.smcontainers import ShakeMapOutputContainer
from esi_utils_rupture.point_rupture import PointRupture

# local imports
from esi_shakelib.utils.containers import ShakeMapInputContainer

homedir = os.path.dirname(os.path.abspath(__file__))  # where is this script?
shakedir = os.path.abspath(os.path.join(homedir, "..", ".."))
sys.path.insert(0, shakedir)


def randomword(length):
    letters = string.ascii_lowercase
    return "".join(random.choice(letters) for i in range(length))


def dict_equal(d1, d2):
    s1 = sorted(set(d1.keys()))
    s2 = sorted(set(d2.keys()))
    return s1 == s2


def test_input_container():
    f, datafile = tempfile.mkstemp()
    os.close(f)
    event_text = """<?xml version="1.0" encoding="US-ASCII" standalone="yes"?>
<earthquake id="2008ryan" lat="30.9858" lon="103.3639" mag="7.9"
time="2008-05-12T06:28:01Z"
depth="19.0" locstring="EASTERN SICHUAN, CHINA" productcode="us2008ryan"
mech="" netid="us" network="" />"""
    try:
        config = {
            "alliance": "chaotic neutral",
            "race": "Elf",
            "armor": 5,
            "class": "Warrior",
            "intelligence": 10,
            "data": {"min_nresp": 3},
        }
        rupturefile = os.path.join(
            homedir, "container_data", "Barkaetal02_fault.txt"
        )
        eventfile = io.StringIO(event_text)
        datafiles = [
            os.path.join(homedir, "container_data/northridge_stations_dat.xml")
        ]

        timestamp = (
            datetime.datetime.fromtimestamp(
                time.time(), datetime.UTC
            ).strftime("%FT%TZ"),
        )
        originator = "us"
        version = 1
        history = {"history": [[timestamp, originator, version]]}

        container = ShakeMapInputContainer.createFromInput(
            datafile,
            config,
            eventfile,
            history,
            datafiles=datafiles,
            rupturefile=rupturefile,
        )
        cfile = container.getFileName()
        assert datafile == cfile
        config = container.getConfig()
        station = container.getStationList()
        rupture = container.getRuptureObject()
        history = container.getVersionHistory()
        container.close()

        container2 = ShakeMapInputContainer.load(datafile)
        config2 = container2.getConfig()
        station2 = container2.getStationList()  # noqa
        rupture2 = container2.getRuptureObject()  # noqa
        history2 = container2.getVersionHistory()  # noqa

        assert dict_equal(config, config2)
        df1, _ = station.getStationDictionary(instrumented=False)
        df2, _ = station2.getStationDictionary(instrumented=False)
        assert dict_equal(df1, df2)
        df1, _ = station.getStationDictionary(instrumented=True)
        df2, _ = station2.getStationDictionary(instrumented=True)
        assert dict_equal(df1, df2)
        assert history["history"][-1][0] == history["history"][-1][0]
        assert history["history"][-1][1] == history["history"][-1][1]
        assert history["history"][-1][2] == history["history"][-1][2]

        container2.close()

        eventfile.seek(0)
        container3 = ShakeMapInputContainer.createFromInput(
            datafile, config, eventfile, {}
        )
        try:
            # this should fail, because we haven't set any station data yet
            station = container3.getStationList()
        except AttributeError:
            assert 1 == 1
        rupture = container3.getRuptureObject()
        history = container3.getVersionHistory()
        assert len(history) == 0
        assert isinstance(rupture, PointRupture)

        container3.setStationData(datafiles)

        #
        # Test the getStationDict() and setStationDict() functions with
        # some dummy data
        #
        config = {
            "alliance": "chaotic neutral",
            "race": "Elf",
            "armor": 5,
            "class": "Warrior",
            "intelligence": 10,
        }
        with pytest.raises(AttributeError):
            container3.getStationDict()
        with pytest.raises(TypeError):
            container3.setStationDict(None)
        container3.setStationDict(config)
        config2 = container3.getStationDict()
        assert dict_equal(config, config2)
        container3.close()

    except Exception as e:
        raise (e)
    finally:
        os.remove(datafile)


def test_output_container():
    nrows = 400
    ncols = 230
    # create MMI mean data for maximum component
    mean_mmi_maximum_data = np.random.rand(nrows, ncols)
    mean_mmi_maximum_metadata = {
        "xmin": -118.5,
        "xmax": -114.5,
        "ymin": 32.1,
        "ymax": 36.7,
        "dx": 0.01,
        "dy": 0.02,
        "nx": 400,
        "ny": 230,
        "name": "Gandalf",
        "color": "white",
        "powers": "magic",
    }

    # create MMI std data for maximum component
    std_mmi_maximum_data = mean_mmi_maximum_data / 10
    std_mmi_maximum_metadata = {
        "xmin": -118.5,
        "xmax": -114.5,
        "ymin": 32.1,
        "ymax": 36.7,
        "dx": 0.01,
        "dy": 0.02,
        "nx": 400,
        "ny": 230,
        "name": "Legolas",
        "color": "green",
        "powers": "good hair",
    }

    # create MMI mean data for rotd50 component
    mean_mmi_rotd50_data = np.random.rand(nrows, ncols)
    mean_mmi_rotd50_metadata = {
        "xmin": -118.5,
        "xmax": -114.5,
        "ymin": 32.1,
        "ymax": 36.7,
        "dx": 0.01,
        "dy": 0.02,
        "nx": 400,
        "ny": 230,
        "name": "Gimli",
        "color": "brown",
        "powers": "axing",
    }

    # create MMI std data for rotd50 component
    std_mmi_rotd50_data = mean_mmi_rotd50_data / 10
    std_mmi_rotd50_metadata = {
        "xmin": -118.5,
        "xmax": -114.5,
        "ymin": 32.1,
        "ymax": 36.7,
        "dx": 0.01,
        "dy": 0.02,
        "nx": 400,
        "ny": 230,
        "name": "Aragorn",
        "color": "white",
        "powers": "scruffiness",
    }

    # create PGA mean data for maximum component
    mean_pga_maximum_data = np.random.rand(nrows, ncols)
    mean_pga_maximum_metadata = {
        "xmin": -118.5,
        "xmax": -114.5,
        "ymin": 32.1,
        "ymax": 36.7,
        "dx": 0.01,
        "dy": 0.02,
        "nx": 400,
        "ny": 230,
        "name": "Pippin",
        "color": "purple",
        "powers": "rashness",
    }

    # create PGA std data for maximum component
    std_pga_maximum_data = mean_pga_maximum_data / 10
    std_pga_maximum_metadata = {
        "xmin": -118.5,
        "xmax": -114.5,
        "ymin": 32.1,
        "ymax": 36.7,
        "dx": 0.01,
        "dy": 0.02,
        "nx": 400,
        "ny": 230,
        "name": "Merry",
        "color": "grey",
        "powers": "hunger",
    }

    f, datafile = tempfile.mkstemp()
    os.close(f)

    try:
        container = ShakeMapOutputContainer.create(datafile)
        # LookupError raised if trying to dropIMTs if there are none
        with pytest.raises(LookupError):
            container.dropIMT("mmi")

        # Add imts
        container.setIMTGrids(
            "mmi",
            "maximum",
            mean_mmi_maximum_data,
            mean_mmi_maximum_metadata,
            std_mmi_maximum_data,
            std_mmi_maximum_metadata,
        )
        container.setIMTGrids(
            "mmi",
            "rotd50",
            mean_mmi_rotd50_data,
            mean_mmi_rotd50_metadata,
            std_mmi_rotd50_data,
            std_mmi_rotd50_metadata,
        )
        container.setIMTGrids(
            "pga",
            "maximum",
            mean_pga_maximum_data,
            mean_pga_maximum_metadata,
            std_pga_maximum_data,
            std_pga_maximum_metadata,
        )

        # get the maximum MMI imt data
        mmi_max_dict = container.getIMTGrids("mmi", component="maximum")
        np.testing.assert_allclose(mmi_max_dict["mean"], mean_mmi_maximum_data)
        np.testing.assert_allclose(mmi_max_dict["std"], std_mmi_maximum_data)
        assert mmi_max_dict["mean_metadata"] == mean_mmi_maximum_metadata
        assert mmi_max_dict["std_metadata"] == std_mmi_maximum_metadata

        # get the rotd50 MMI imt data
        mmi_rot_dict = container.getIMTGrids("mmi", component="rotd50")
        np.testing.assert_allclose(mmi_rot_dict["mean"], mean_mmi_rotd50_data)
        np.testing.assert_allclose(mmi_rot_dict["std"], std_mmi_rotd50_data)
        assert mmi_rot_dict["mean_metadata"] == mean_mmi_rotd50_metadata
        assert mmi_rot_dict["std_metadata"] == std_mmi_rotd50_metadata

        # get list of all imts
        imts = container.getIMTs()

        # get list of maximum imts
        max_imts = container.getIMTs(component="maximum")
        assert sorted(max_imts) == ["mmi", "pga"]

        # get list of components for mmi
        mmi_comps = container.getComponents("mmi")
        assert sorted(mmi_comps) == ["maximum", "rotd50"]

        # Test dropIMT
        imts = container.getIMTs("maximum")
        assert imts == ["mmi", "pga"]
        container.dropIMT("mmi")
        imts = container.getIMTs("maximum")
        assert imts == ["pga"]
        container.close()

    except Exception as e:
        raise (e)
    finally:
        os.remove(datafile)


def test_output_arrays():

    f, datafile = tempfile.mkstemp()
    os.close(f)

    try:
        container = ShakeMapOutputContainer.create(datafile)
        #
        # Test that no data type is set
        #
        assert container.getDataType() is None

        #
        # Make some array data and metadata
        #
        mean = np.random.rand(100)
        std = np.random.rand(100)
        lats = np.random.rand(100)
        lons = np.random.rand(100)
        ids = np.array([randomword(4).encode("ascii") for x in range(100)])
        metadata = {"units": "%g", "digits": 4}
        #
        # Put the data in the container
        #
        container.setIMTArrays(
            "PGA", "Larger", lons, lats, ids, mean, metadata, std, metadata
        )
        #
        # Now extract it and compare it to what we put in there
        #
        dout = container.getIMTArrays("PGA", "Larger")
        np.testing.assert_allclose(dout["lons"], lons)
        np.testing.assert_allclose(dout["lats"], lats)
        assert all(dout["ids"] == ids)
        np.testing.assert_allclose(dout["mean"], mean)
        np.testing.assert_allclose(dout["std"], std)
        #
        # Check the data type
        #
        assert container.getDataType() == "points"
        #
        # Try raising some exceptions
        #
        # Shouldn't be able to find this IMT
        with pytest.raises(LookupError):
            container.getIMTArrays("JUNK", "Larger")
        # Shapes of inputs not the same
        with pytest.raises(ValueError):
            empty = np.array([])
            container.setIMTArrays(
                "PGV",
                "Larger",
                empty,
                lats,
                ids,
                mean,
                metadata,
                std,
                metadata,
            )
        # IMT already exists
        with pytest.raises(LookupError):
            container.setIMTArrays(
                "PGA", "Larger", lons, lats, ids, mean, metadata, std, metadata
            )
        # Trying to set a grid in a file with points
        with pytest.raises(TypeError):
            container.setIMTGrids(
                "PGV", "Larger", mean, metadata, std, metadata
            )
        # Trying to get a grid in a file with points
        with pytest.raises(TypeError):
            container.getIMTGrids("PGA", "Larger")

        container.close()

    except Exception as e:
        if os.path.isfile(datafile):
            os.remove(datafile)
        raise (e)
    finally:
        if os.path.isfile(datafile):
            os.remove(datafile)


if __name__ == "__main__":
    test_input_container()
    test_output_container()
    test_output_arrays()



"""
Module to implement CB14 basin and M9 modifications to Atkinson and Macias (2009) GSIM.
"""

from esi_shakelib.gmpe.modifiable_gmpes.m9_basin import M9BasinTerm


class AtkinsonMacias2009basin(M9BasinTerm):
    """
    Returns basin modification of the Atkinson and Macias (2009) GSIM.
    """

    m9 = False
    SJ = 0

    def __init__(self):
        super().__init__(gmpe_name="AtkinsonMacias2009")


class AtkinsonMacias2009M9basin(M9BasinTerm):
    """
    Returns M9+basin modification of the Atkinson and Macias (2009) GSIM.
    """

    m9 = True
    SJ = 0

    def __init__(self):
        super().__init__(gmpe_name="AtkinsonMacias2009")


class AtkinsonMacias2009Japanbasin(M9BasinTerm):
    """
    Returns Japan basin modification of the Atkinson and Macias (2009) GSIM.
    """

    m9 = False
    SJ = 1

    def __init__(self):
        super().__init__(gmpe_name="AtkinsonMacias2009")


class AtkinsonMacias2009JapanM9basin(M9BasinTerm):
    """
    Returns Japan M9+basin modification of the Atkinson and Macias (2009) GSIM.
    """

    m9 = True
    SJ = 1

    def __init__(self):
        super().__init__(gmpe_name="AtkinsonMacias2009")
